#include <iostream>
#include <omp.h>
#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>
#include <torch/extension.h>
#include <time.h>
#include<random>
#include <parallel_hashmap/phmap.h>

using namespace std;
namespace py = pybind11;
namespace th = torch;

typedef int64_t NodeIDType;
// typedef int64_t EdgeIDType;
// typedef float TimeStampType;

class TemporalNeighborBlock;
class TemporalGraphBlock;
TemporalNeighborBlock get_neighbors(th::Tensor row, th::Tensor col, int64_t num_nodes);
TemporalGraphBlock* neighbor_sample_from_node(NodeIDType node, vector<NodeIDType> neighbors, int64_t deg, int64_t fanout, int threads);
vector<th::Tensor> neighbor_sample_from_nodes(th::Tensor nodes, TemporalNeighborBlock& tnb, int pid, const vector<NodeIDType>& part_ptr, int64_t fanout, int threads);
th::Tensor heads_unique(th::Tensor array, th::Tensor heads, int threads);
int nodeIdToInOut(NodeIDType nid, int pid, const vector<NodeIDType>& part_ptr);
int nodeIdToPartId(NodeIDType nid, const vector<NodeIDType>& part_ptr);


template<typename T>
inline py::array vec2npy(const std::vector<T> &vec)
{
    // need to let python garbage collector handle C++ vector memory 
    // see https://github.com/pybind/pybind11/issues/1042
    // non-copy value transfer
    auto v = new std::vector<T>(vec);
    auto capsule = py::capsule(v, [](void *v)
                               { delete reinterpret_cast<std::vector<T> *>(v); });
    return py::array(v->size(), v->data(), capsule);
    // return py::array(vec.size(), vec.data());
}

/* 
 * NeighborSampler Utils
 */
class TemporalNeighborBlock
{
    public:
        std::vector<vector<NodeIDType>> neighbors;
        std::vector<int64_t> deg;
        // int64_t max_deg;

        TemporalNeighborBlock(){}
        TemporalNeighborBlock(const TemporalNeighborBlock &tnb);
        TemporalNeighborBlock(std::vector<vector<NodeIDType>>& neighbors, 
                              std::vector<int64_t> &deg):
                              neighbors(neighbors), deg(deg){}
        py::array get_node_neighbor(NodeIDType node_id){
            return vec2npy(neighbors[node_id]);
        }
        int64_t get_node_deg(NodeIDType node_id){
            return deg[node_id];
        }
        // int64_t get_max_deg(){
        //     return max_deg;
        // }
};

//캯
TemporalNeighborBlock::TemporalNeighborBlock(const TemporalNeighborBlock &tnb){
    // clock_t pre = clock();
    this->neighbors = tnb.neighbors;
    this->deg = tnb.deg;
    // this->max_deg = tnb.max_deg;
    // clock_t end=clock();
    // cout<<"copy tnb time: "<<(end-pre)*1.0/CLOCKS_PER_SEC<<endl;
}

class TemporalGraphBlock
{
    public:
        std::vector<NodeIDType> row;
        std::vector<NodeIDType> col;
        std::vector<NodeIDType> inner_nodes;
        std::vector<NodeIDType> outer_nodes;

        TemporalGraphBlock(){}
        TemporalGraphBlock(const TemporalGraphBlock &tgb);
        TemporalGraphBlock(std::vector<NodeIDType> &_row, std::vector<NodeIDType> &_col,
                           std::vector<NodeIDType> &_inner_nodes, std::vector<NodeIDType> &_outer_nodes):
                           row(_row), col(_col), inner_nodes(_inner_nodes), outer_nodes(_outer_nodes){}
};

//캯
TemporalGraphBlock::TemporalGraphBlock(const TemporalGraphBlock &tgb){
    // clock_t pre = clock();
    this->row = tgb.row;
    this->col = tgb.col;
    this->inner_nodes = tgb.inner_nodes;
    this->outer_nodes = tgb.outer_nodes;
    // clock_t end=clock();
    // cout<<"copy tgb time: "<<(end-pre)*1.0/CLOCKS_PER_SEC<<endl;
}

TemporalNeighborBlock get_neighbors(
        th::Tensor row, th::Tensor col, int64_t num_nodes){
    AT_ASSERTM(row.is_contiguous(), "Offset tensor must be contiguous");
    AT_ASSERTM(col.is_contiguous(), "Offset tensor must be contiguous");
    AT_ASSERTM(row.dim() == 1, "0ffset tensor must be one-dimensional");
    AT_ASSERTM(col.dim() == 1, "0ffset tensor must be one-dimensional");
    auto u = row.data_ptr<NodeIDType>();
    auto v = col.data_ptr<NodeIDType>();
    int64_t edge_num = row.size(0);
    TemporalNeighborBlock tnb = TemporalNeighborBlock();
    tnb.deg.resize(num_nodes, 0);
    // tnb.max_deg=0;
    double start_time = omp_get_wtime();
    for(int64_t i=0; i<num_nodes; i++)
        tnb.neighbors.push_back(vector<NodeIDType>());
    for(int64_t i=0; i<edge_num; i++){
        //ڵھ
        tnb.neighbors[u[i]].push_back(v[i]);
    }
    for(int64_t i=0; i<num_nodes; i++){
        //ռڵ
        tnb.deg[i] = int64_t(tnb.neighbors[i].size());
        // tnb.max_deg = max(tnb.max_deg, tnb.deg[i]);
    }
    double end_time = omp_get_wtime();
    cout<<"get_neighbors consume: "<<end_time-start_time<<"s"<<endl;
    return tnb;
}

vector<th::Tensor> neighbor_sample_from_nodes(
        th::Tensor nodes, TemporalNeighborBlock& tnb, int pid, const vector<NodeIDType>& part_ptr, int64_t fanout, int threads){
    py::gil_scoped_release release;
    TemporalGraphBlock tgb = TemporalGraphBlock();
    vector<th::Tensor> ret;
    AT_ASSERTM(nodes.is_contiguous(), "Offset tensor must be contiguous");
    AT_ASSERTM(nodes.dim() == 1, "0ffset tensor must be one-dimensional");
    auto nodes_data = nodes.data_ptr<NodeIDType>();
    vector<phmap::parallel_flat_hash_set<NodeIDType>> inner_s_threads;
    vector<phmap::parallel_flat_hash_set<NodeIDType>> outer_s_threads;
    phmap::parallel_flat_hash_set<NodeIDType> inner_s;
    phmap::parallel_flat_hash_set<NodeIDType> outer_s;
    vector<vector<NodeIDType>> row_threads, col_threads;
    
    for(int i = 0; i<threads; i++){
        row_threads.push_back(vector<NodeIDType>());
        col_threads.push_back(vector<NodeIDType>());
        inner_s_threads.push_back(phmap::parallel_flat_hash_set<NodeIDType>());
        outer_s_threads.push_back(phmap::parallel_flat_hash_set<NodeIDType>());
    }
    double start_time = omp_get_wtime();
#pragma omp parallel for num_threads(threads) default(shared)
    for(int64_t i=0; i<nodes.size(0); i++){
        int tid = omp_get_thread_num();
        NodeIDType node = nodes_data[i];
        vector<NodeIDType> nei(tnb.neighbors[node]);
        TemporalGraphBlock tgb_i = TemporalGraphBlock();
        default_random_engine e(time(0));
        uniform_int_distribution<> u(0, tnb.deg[node]-1);
        if(tnb.deg[node]>fanout){
            //ȴȳĻҪѡfanoutھ
            phmap::flat_hash_set<NodeIDType> temp_s;
            while(temp_s.size()!=fanout){
                //ѭѡfanoutھ
                auto chosen_iter = nei.begin() + u(e);
                if(nodeIdToInOut(*chosen_iter, pid, part_ptr)==0)
                    inner_s_threads[tid].insert(*chosen_iter);
                else
                    outer_s_threads[tid].insert(*chosen_iter);
                temp_s.insert(*chosen_iter);
            }
            tgb_i.col.assign(temp_s.begin(), temp_s.end());
        }
        else{
            for(const auto& node:nei){
                if(nodeIdToInOut(node, pid, part_ptr)==0)
                    inner_s_threads[tid].insert(node);
                else
                    outer_s_threads[tid].insert(node);
            }
            tgb_i.col.swap(nei);
        }
        tgb_i.row.resize(tgb_i.col.size(), node);
        // TemporalGraphBlock* tgb_i = neighbor_sample_from_node(node, tnb.neighbors[node], tnb.deg[node], fanout, threads);
        row_threads[tid].insert(row_threads[tid].end(),tgb_i.row.begin(),tgb_i.row.end());
        col_threads[tid].insert(col_threads[tid].end(),tgb_i.col.begin(),tgb_i.col.end());
    }
    double end_time = omp_get_wtime();
    cout<<"neighbor_sample_from_nodes parallel part consume: "<<end_time-start_time<<"s"<<endl;
#pragma omp sections
    {
#pragma omp section
        for(int i = 0; i<threads; i++)
            tgb.row.insert(tgb.row.end(), row_threads[i].begin(), row_threads[i].end());
#pragma omp section
        for(int i = 0; i<threads; i++)
            tgb.col.insert(tgb.col.end(), col_threads[i].begin(), col_threads[i].end());
#pragma omp section
        for(int i = 0; i<threads; i++)
            inner_s.insert(inner_s_threads[i].begin(), inner_s_threads[i].end());
#pragma omp section
        for(int i = 0; i<threads; i++)
            outer_s.insert(outer_s_threads[i].begin(), outer_s_threads[i].end());
    }
    //sampled nodes ȥ, ݲroot nodes
    start_time = end_time;
    // phmap::parallel_flat_hash_set<NodeIDType> s(tgb.col.begin(), tgb.col.end());
    tgb.inner_nodes.assign(inner_s.begin(), inner_s.end());
    tgb.outer_nodes.assign(outer_s.begin(), outer_s.end());
    end_time = omp_get_wtime();
    cout<<"end unique consume: "<<end_time-start_time<<"s"<<endl;

    ret.push_back(th::tensor(tgb.row));
    ret.push_back(th::tensor(tgb.col));
    ret.push_back(th::tensor(tgb.inner_nodes));
    ret.push_back(th::tensor(tgb.outer_nodes));

    py::gil_scoped_acquire acquire;
    return ret;
}

/*-------------------------------------------------------------------------------------**
**------------Utils--------------------------------------------------------------------**
**-------------------------------------------------------------------------------------*/

th::Tensor heads_unique(th::Tensor array, th::Tensor heads, int threads){
    auto array_ptr = array.data_ptr<NodeIDType>();
    phmap::parallel_flat_hash_set<NodeIDType> s(array_ptr, array_ptr+array.numel());
    if(heads.numel()==0) return th::tensor(vector<NodeIDType>(s.begin(), s.end()));
    AT_ASSERTM(heads.is_contiguous(), "Offset tensor must be contiguous");
    AT_ASSERTM(heads.dim() == 1, "0ffset tensor must be one-dimensional");
    auto heads_ptr = heads.data_ptr<NodeIDType>();
#pragma omp parallel for num_threads(threads)
    for(int64_t i=0; i<heads.size(0); i++){
        if(s.count(heads_ptr[i])==1){
        #pragma omp critical(erase)
            s.erase(heads_ptr[i]);
        }
    }
    vector<NodeIDType> ret;
    ret.reserve(s.size()+heads.numel());
    ret.assign(heads_ptr, heads_ptr+heads.numel());
    ret.insert(ret.end(), s.begin(), s.end());
    // cout<<"s: "<<s.size()<<" array: "<<array.size()<<endl;
    return th::tensor(ret);
}

int nodeIdToPartId(NodeIDType nid, const vector<NodeIDType>& part_ptr){
    int partitionId = -1;
    for(int i=0;i<part_ptr.size();i++){
            if(nid>=part_ptr[i]&&nid<part_ptr[i+1]){
                partitionId = i;
                break;
            }
    }
    if(partitionId<0) throw "nid ڶӦķ";
    return partitionId;
}
//0:inner; 1:outer
int nodeIdToInOut(NodeIDType nid, int pid, const vector<NodeIDType>& part_ptr){
    if(nid>=part_ptr[pid]&&nid<part_ptr[pid+1]){
        return 0;
    }
    return 1;
}

/*------------Python Bind--------------------------------------------------------------*/
PYBIND11_MODULE(sample_cores, m)
{
    m
    .def("neighbor_sample_from_nodes", 
        &neighbor_sample_from_nodes, 
        py::return_value_policy::reference)
    .def("get_neighbors", 
        &get_neighbors, 
        py::return_value_policy::reference)
    .def("heads_unique", 
        &heads_unique, 
        py::return_value_policy::reference);

    py::class_<TemporalGraphBlock>(m, "TemporalGraphBlock")
        .def(py::init<std::vector<NodeIDType> &, std::vector<NodeIDType> &,
                      std::vector<NodeIDType> &, std::vector<NodeIDType> &>())
        .def("row", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.row); })
        .def("col", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.col); })
        .def("inner_nodes", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.inner_nodes); })
        .def("outer_nodes", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.outer_nodes); });

    py::class_<TemporalNeighborBlock>(m, "TemporalNeighborBlock")
        .def(py::init<std::vector<vector<NodeIDType>>&, 
                      std::vector<int64_t> &>())
        // .def("get_node_neighbor",&TemporalNeighborBlock::get_node_neighbor)
        // .def("get_node_deg", &TemporalNeighborBlock::get_node_deg)
        // .def("get_max_deg", &TemporalNeighborBlock::get_max_deg)
        // .def_readonly("max_deg", &TemporalNeighborBlock::max_deg, py::return_value_policy::reference)
        .def_readonly("neighbors", &TemporalNeighborBlock::neighbors, py::return_value_policy::reference)
        .def_readonly("deg", &TemporalNeighborBlock::deg, py::return_value_policy::reference);
}